# Copyright (c) OpenMMLab. All rights reserved.
import mmcv
import numpy as np
import torch

from mmagic.utils import reorder_image


def _assert_ndim(input, name, ndim, shape_hint):
    if input.ndim != ndim:
        raise ValueError(
            f'{name} should be of shape {shape_hint}, but got {input.shape}.')


def _assert_masked(pred_alpha, trimap):
    if (pred_alpha[trimap == 0] != 0).any() or (pred_alpha[trimap == 255] !=
                                                255).any():
        raise ValueError(
            'pred_alpha should be masked by trimap before evaluation')


def _fetch_data_and_check(data_samples):
    """Fetch and check data from one item of data_batch and predictions.

    Args:
        data_batch (dict): One item of data_batch.
        predictions (dict): One item of predictions.

    Returns:
        pred_alpha (Tensor): Pred_alpha data of predictions.
        ori_alpha (Tensor): Ori_alpha data of data_batch.
        ori_trimap (Tensor): Ori_trimap data of data_batch.
    """
    ori_trimap = data_samples['ori_trimap'][0, :, :].cpu().numpy()
    ori_alpha = data_samples['ori_alpha'][0, :, :].cpu().numpy()
    pred_alpha = data_samples['output']['pred_alpha']  # 2D tensor
    pred_alpha = pred_alpha.cpu().numpy()

    _assert_ndim(ori_trimap, 'trimap', 2, 'HxW')
    _assert_ndim(ori_alpha, 'gt_alpha', 2, 'HxW')
    _assert_ndim(pred_alpha, 'pred_alpha', 2, 'HxW')
    _assert_masked(pred_alpha, ori_trimap)

    # dtype uint8 -> float64
    pred_alpha = pred_alpha / 255.0
    ori_alpha = ori_alpha / 255.0
    # test shows that using float32 vs float64 differs final results at 1e-4
    # speed are comparable, so we choose float64 for accuracy

    return pred_alpha, ori_alpha, ori_trimap


def average(results, key):
    """Average of key in results(list[dict]).

    Args:
        results (list[dict]): A list of dict containing the necessary data.
        key (str): The key of target data.

    Returns:
        result: The average result.
    """

    total = 0
    n = 0
    for batch_result in results:
        batch_size = batch_result.get('batch_size', 1)
        total += batch_result[key] * batch_size
        n += batch_size

    return total / n


def img_transform(img,
                  crop_border=0,
                  input_order='HWC',
                  convert_to=None,
                  channel_order='rgb'):
    """Image transform.

    Ref: https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio

    Args:
        img (np.ndarray): Images with range [0, 255].
        crop_border (int): Cropped pixels in each edges of an image. These
            pixels are not involved in the PSNR calculation. Default: 0.
        input_order (str): Whether the input order is 'HWC' or 'CHW'.
            Default: 'HWC'.
        convert_to (str): Whether to convert the images to other color models.
            If None, the images are not altered. When computing for 'Y',
            the images are assumed to be in BGR order. Options are 'Y' and
            None. Default: None.
        channel_order (str): The channel order of image. Default: 'rgb'

    Returns:
        float: PSNR result.
    """

    if input_order not in ['HWC', 'CHW']:
        raise ValueError(
            f'Wrong input_order {input_order}. Supported input_orders are '
            '"HWC" and "CHW"')

    img = reorder_image(img, input_order=input_order)
    if isinstance(img, torch.Tensor):
        img = img.numpy()
    img = img.astype(np.float32)

    if isinstance(convert_to, str) and convert_to.lower() == 'y':
        if channel_order.upper() == 'RGB':
            img = mmcv.rgb2ycbcr(img / 255., y_only=True) * 255.
        elif channel_order.upper() == 'BGR':
            img = mmcv.bgr2ycbcr(img / 255., y_only=True) * 255.
        else:
            raise ValueError(
                'Only support `rgb2y` and `bgr2`, but the channel_order '
                f'is {channel_order}')
        img = np.expand_dims(img, axis=2)
    elif convert_to is not None:
        raise ValueError('Wrong color model. Supported values are '
                         '"Y" and None.')

    if crop_border != 0:
        img = img[crop_border:-crop_border, crop_border:-crop_border, ...]

    return img


def obtain_data(data_sample, key, device='cpu'):
    """Obtain data of key from data_sample and converse data to device.
    Args:
        data_sample (dict): A dict of data sample.
        key (str): The key of data to obtain.
        device (str): Which device the data will deploy. Default: 'cpu'.

    Returns:
        result (Tensor | np.ndarray): The data of key.
    """
    candidates = ['data_samples', key, 'data']

    for k in candidates:
        if k in data_sample:
            result = data_sample[k]
            if isinstance(result, dict):
                return obtain_data(result, key, device)
            else:
                if isinstance(result, torch.Tensor):
                    result = result.to(device)
                return result

    raise KeyError('Mapping key was not found')



# Copyright (c) OpenMMLab. All rights reserved.
import inspect
from copy import deepcopy
from typing import Dict, List, Optional, Union

import torch
import torch.nn as nn
from einops import rearrange
from mmengine import print_log
from mmengine.logging import MMLogger
from mmengine.model import BaseModel
from safetensors import safe_open
from tqdm import tqdm

from mmagic.models.archs import TokenizerWrapper, set_lora
from mmagic.models.utils import build_module, set_tomesd, set_xformers
from mmagic.registry import DIFFUSION_SCHEDULERS, MODELS
from mmagic.structures import DataSample
from mmagic.utils.typing import SampleList
from .animatediff_utils import (convert_ldm_clip_checkpoint,
                                convert_ldm_unet_checkpoint,
                                convert_ldm_vae_checkpoint)

logger = MMLogger.get_current_instance()

ModelType = Union[Dict, nn.Module]


@MODELS.register_module('animatediff')
@MODELS.register_module()
class AnimateDiff(BaseModel):
    """Implementation of `AnimateDiff.

    <https://arxiv.org/abs/2307.04725>`_ (AnimateDiff).

    Args:
        vae (Union[dict, nn.Module]): The config or module for VAE model.
        text_encoder (Union[dict, nn.Module]): The config or module for text
            encoder.
        tokenizer (str): The **name** for CLIP tokenizer.
        unet (Union[dict, nn.Module]): The config or module for Unet model.
        schedule (Union[dict, nn.Module]): The config or module for diffusion
            scheduler.
        test_scheduler (Union[dict, nn.Module], optional): The config or
            module for diffusion scheduler in test stage (`self.infer`). If not
            passed, will use the same scheduler as `schedule`. Defaults to
            None.
        lora_config (dict, optional): The config for LoRA finetuning. Defaults
            to None.
        val_prompts (Union[str, List[str]], optional): The prompts for
            validation. Defaults to None.
        class_prior_prompt (str, optional): The prompt for class prior loss.
        num_class_images (int, optional): The number of images for class prior.
            Defaults to 3.
        prior_loss_weight (float, optional): The weight for class prior loss.
            Defaults to 0.
        fine_tune_text_encoder (bool, optional): Whether to fine-tune text
            encoder. Defaults to False.
        dtype (str, optional): The dtype for the model. Defaults to 'fp16'.
        enable_xformers (bool, optional): Whether to use xformers.
            Defaults to True.
        noise_offset_weight (bool, optional): The weight of noise offset
            introduced in https://www.crosslabs.org/blog/diffusion-with-offset-noise  # noqa
            Defaults to 0.
        tomesd_cfg (dict, optional): The config for TOMESD. Please refers to
            https://github.com/dbolya/tomesd and
            https://github.com/open-mmlab/mmagic/blob/main/mmagic/models/utils/tome_utils.py for detail.  # noqa
            Defaults to None.
        data_preprocessor (dict, optional): The pre-process config of
            :class:`BaseDataPreprocessor`. Defaults to
                dict(type='DataPreprocessor').
        init_cfg (dict, optional): The weight initialized config for
            :class:`BaseModule`. Defaults to None/
    """

    def __init__(
        self,
        vae: ModelType,
        text_encoder: ModelType,
        tokenizer: str,
        unet: ModelType,
        scheduler: ModelType,
        test_scheduler: Optional[ModelType] = None,
        dtype: str = 'fp32',
        enable_xformers: bool = True,
        noise_offset_weight: float = 0,
        tomesd_cfg: Optional[dict] = None,
        data_preprocessor=dict(type='DataPreprocessor'),
        motion_module_cfg: Optional[dict] = None,
        dream_booth_lora_cfg: Optional[dict] = None,
    ):
        super().__init__(data_preprocessor)

        default_args = dict()
        if dtype is not None:
            default_args['dtype'] = dtype

        self.dtype = torch.float32
        if dtype in ['float16', 'fp16', 'half']:
            self.dtype = torch.float16
        elif dtype == 'bf16':
            self.dtype = torch.bfloat16
        else:
            assert dtype in [
                'fp32', None
            ], ('dtype must be one of \'fp32\', \'fp16\', \'bf16\' or None.')
        self.vae = build_module(vae, MODELS, default_args=default_args)
        self.unet = build_module(unet, MODELS)  # NOTE: initialize unet as fp32
        self._unet_ori_dtype = next(self.unet.parameters()).dtype
        print_log(f'Set UNet dtype to \'{self._unet_ori_dtype}\'.', 'current')

        self.init_motion_module(motion_module_cfg)

        self.scheduler = build_module(scheduler, DIFFUSION_SCHEDULERS)
        if test_scheduler is None:
            self.test_scheduler = deepcopy(self.scheduler)
        else:
            self.test_scheduler = build_module(test_scheduler,
                                               DIFFUSION_SCHEDULERS)
        self.text_encoder = build_module(text_encoder, MODELS)
        if not isinstance(tokenizer, str):
            self.tokenizer = tokenizer
        else:
            # NOTE: here we assume tokenizer is an string
            self.tokenizer = TokenizerWrapper(tokenizer, subfolder='tokenizer')

        self.unet_sample_size = self.unet.sample_size
        self.vae_scale_factor = 2**(len(self.vae.block_out_channels) - 1)

        self.enable_noise_offset = noise_offset_weight > 0
        self.noise_offset_weight = noise_offset_weight

        self.enable_xformers = enable_xformers
        self.unet.set_use_memory_efficient_attention_xformers(True)

        self.tomesd_cfg = tomesd_cfg
        self.set_tomesd()

        self.init_dreambooth_lora(dream_booth_lora_cfg)

        self.prepare_model()

    def set_xformers(self, module: Optional[nn.Module] = None) -> nn.Module:
        """Set xformers for the model.

        Returns:
            nn.Module: The model with xformers.
        """
        if self.enable_xformers:
            if module is None:
                set_xformers(self)
            else:
                set_xformers(module)

    def set_tomesd(self) -> nn.Module:
        """Set ToMe for the stable diffusion model.

        Returns:
            nn.Module: The model with ToMe.
        """
        if self.tomesd_cfg is not None:
            set_tomesd(self, **self.tomesd_cfg)

    @property
    def device(self):
        """Set device for the model."""
        return next(self.parameters()).device

    def init_motion_module(self, motion_module_cfg):
        if motion_module_cfg is not None:
            if 'path' in motion_module_cfg.keys():
                motion_module_state_dict = torch.load(
                    motion_module_cfg['path'], map_location='cpu')
                # if "global_step" in motion_module_state_dict:
                # func_args.update({"global_step":
                # motion_module_state_dict["global_step"]})
                missing, unexpected = self.unet.load_state_dict(
                    motion_module_state_dict, strict=False)
                assert len(unexpected) == 0

    def init_dreambooth_lora(self, dream_booth_lora_cfg):
        # TODO: finish
        if dream_booth_lora_cfg is not None:
            if 'path' in dream_booth_lora_cfg.keys():
                state_dict = {}
                with safe_open(
                        dream_booth_lora_cfg['path'], framework='pt',
                        device='cpu') as f:
                    for key in f.keys():
                        state_dict[key] = f.get_tensor(key)
                # vae
                converted_vae_checkpoint = convert_ldm_vae_checkpoint(
                    state_dict, self.vae.config)
                self.vae.load_state_dict(converted_vae_checkpoint)
                # unet
                converted_unet_checkpoint = convert_ldm_unet_checkpoint(
                    state_dict, self.unet.config)
                self.unet.load_state_dict(
                    converted_unet_checkpoint, strict=False)
                # text_model
                self.text_encoder = convert_ldm_clip_checkpoint(state_dict)
                # self.convert_lora(state_dict)

    def _encode_prompt(self, prompt, device, num_videos_per_prompt,
                       do_classifier_free_guidance, negative_prompt):
        """Encodes the prompt into text encoder hidden states."""
        batch_size = len(prompt) if isinstance(prompt, list) else 1

        text_inputs = self.tokenizer(
            prompt,
            padding='max_length',
            max_length=self.tokenizer.model_max_length,
            truncation=True,
            return_tensors='pt',
        )
        text_input_ids = text_inputs.input_ids
        untruncated_ids = self.tokenizer(
            prompt, padding='longest', return_tensors='pt').input_ids

        if untruncated_ids.shape[-1] >= text_input_ids.shape[
                -1] and not torch.equal(text_input_ids, untruncated_ids):
            removed_text = self.tokenizer.batch_decode(
                untruncated_ids[:, self.tokenizer.model_max_length - 1:-1])
            logger.warning(
                'The following part of your input was truncated '
                f'because CLIP can only handle sequences up to'
                f' {self.tokenizer.model_max_length} tokens: {removed_text}')

        text_encoder = self.text_encoder.module if hasattr(
            self.text_encoder, 'module') else self.text_encoder
        if hasattr(text_encoder.config, 'use_attention_mask'
                   ) and text_encoder.config.use_attention_mask:
            attention_mask = text_inputs.attention_mask.to(device)
        else:
            attention_mask = None

        text_embeddings = self.text_encoder(
            text_input_ids.to(device),
            attention_mask=attention_mask,
        )
        text_embeddings = text_embeddings[0]

        # duplicate text embeddings for each generation
        # per prompt, using mps friendly method
        bs_embed, seq_len, _ = text_embeddings.shape
        text_embeddings = text_embeddings.repeat(1, num_videos_per_prompt, 1)
        text_embeddings = text_embeddings.view(
            bs_embed * num_videos_per_prompt, seq_len, -1)

        # get unconditional embeddings for classifier free guidance
        if do_classifier_free_guidance:
            uncond_tokens: List[str]
            if negative_prompt is None:
                uncond_tokens = [''] * batch_size
            elif type(prompt) is not type(negative_prompt):
                raise TypeError(
                    f'`negative_prompt` should be the same type '
                    f'to `prompt`, but got {type(negative_prompt)} !='
                    f' {type(prompt)}.')
            elif isinstance(negative_prompt, str):
                uncond_tokens = [negative_prompt]
            elif batch_size != len(negative_prompt):
                raise ValueError(
                    f'`negative_prompt`: {negative_prompt} has '
                    f'batch size {len(negative_prompt)}, but `prompt`:'
                    f' {prompt} has batch size {batch_size}. Please '
                    f'make sure that passed `negative_prompt` matches'
                    ' the batch size of `prompt`.')
            else:
                uncond_tokens = negative_prompt

            max_length = text_input_ids.shape[-1]
            uncond_input = self.tokenizer(
                uncond_tokens,
                padding='max_length',
                max_length=max_length,
                truncation=True,
                return_tensors='pt',
            )

            if hasattr(text_encoder.config, 'use_attention_mask'
                       ) and text_encoder.config.use_attention_mask:
                attention_mask = uncond_input.attention_mask.to(device)
            else:
                attention_mask = None

            uncond_embeddings = self.text_encoder(
                uncond_input.input_ids.to(device),
                attention_mask=attention_mask,
            )
            uncond_embeddings = uncond_embeddings[0]

            # duplicate unconditional embeddings for each generation
            # per prompt, using mps friendly method
            seq_len = uncond_embeddings.shape[1]
            uncond_embeddings = uncond_embeddings.repeat(
                1, num_videos_per_prompt, 1)
            uncond_embeddings = uncond_embeddings.view(
                batch_size * num_videos_per_prompt, seq_len, -1)

            # For classifier free guidance, we need to do two forward passes.
            # Here we concatenate the unconditional and text embeddings
            # into a single batch to avoid doing two forward passes
            text_embeddings = torch.cat([uncond_embeddings, text_embeddings])

        return text_embeddings

    def decode_latents(self, latents):
        """latents decoder."""
        video_length = latents.shape[2]
        latents = 1 / 0.18215 * latents
        latents = rearrange(latents, 'b c f h w -> (b f) c h w')
        # video = self.vae.decode(latents).sample
        video = []
        for frame_idx in tqdm(range(latents.shape[0])):
            video.append(
                self.vae.decode(latents[frame_idx:frame_idx + 1]).sample)
        video = torch.cat(video)
        video = rearrange(video, '(b f) c h w -> b c f h w', f=video_length)
        video = (video / 2 + 0.5).clamp(0, 1)
        # we always cast to float32 as this does not cause significant
        # overhead and is compatible with bfloa16
        video = video.cpu().float().numpy()
        return video

    def prepare_extra_step_kwargs(self, generator, eta):
        """Prepare extra kwargs for the scheduler step, since not all
        schedulers have the same signature eta (η) is only used with the
        DDIMScheduler, it will be ignored for other schedulers."""
        # prepare extra kwargs for the scheduler step, since not all
        # schedulers have the same signature
        # eta (η) is only used with the DDIMScheduler, it will
        # be ignored for other schedulers.
        # eta corresponds to η in DDIM paper: https://arxiv.org/abs/2010.02502
        # and should be between [0, 1]

        accepts_eta = 'eta' in set(
            inspect.signature(self.scheduler.step).parameters.keys())
        extra_step_kwargs = {}
        if accepts_eta:
            extra_step_kwargs['eta'] = eta

        # check if the scheduler accepts generator
        accepts_generator = 'generator' in set(
            inspect.signature(self.scheduler.step).parameters.keys())
        if accepts_generator:
            extra_step_kwargs['generator'] = generator
        return extra_step_kwargs

    def check_inputs(self, prompt, height, width):
        """Check inputs.

        Raise error if not correct
        """
        if not isinstance(prompt, str) and not isinstance(prompt, list):
            raise ValueError(f'`prompt` has to be of type `str`'
                             f' or `list` but is {type(prompt)}')

        if height % 8 != 0 or width % 8 != 0:
            raise ValueError(f'`height` and `width` have to be divisible'
                             f' by 8 but are {height} and {width}.')

        # if (callback_steps is None) or (
        #     callback_steps is not None and (not isinstance(callback_steps,
        #       int) or callback_steps <= 0)
        # ):
        #     raise ValueError(
        #         f"`callback_steps` has to be a positive integer but
        #         is {callback_steps} of type"
        #         f" {type(callback_steps)}."
        #     )

    def convert_lora(self,
                     state_dict,
                     LORA_PREFIX_UNET='lora_unet',
                     LORA_PREFIX_TEXT_ENCODER='lora_te',
                     alpha=0.6):
        """ Convert lora for unet and text_encoder
            TODO: use this function to convert lora

        Args:
            state_dict (_type_): _description_
            LORA_PREFIX_UNET (str, optional):
            _description_. Defaults to 'lora_unet'.
            LORA_PREFIX_TEXT_ENCODER (str, optional):
            _description_. Defaults to 'lora_te'.
            alpha (float, optional): _description_. Defaults to 0.6.

        Returns:
            TODO: check each output type
            _type_: unet && text_encoder
        """
        # load base model
        # pipeline = StableDiffusionPipeline.from_pretrained(base_model_path,
        # torch_dtype=torch.float32)

        # load LoRA weight from .safetensors
        # state_dict = load_file(checkpoint_path)

        visited = []

        # directly update weight in diffusers model
        for key in state_dict:
            # it is suggested to print out the key, it usually
            # will be something like below
            # "lora_te_text_model_encoder_layers_0_self_attn_k_proj.lora_down.weight"

            # as we have set the alpha beforehand, so just skip
            if '.alpha' in key or key in visited:
                continue

            if 'text' in key:
                layer_infos = key.split('.')[0].split(
                    LORA_PREFIX_TEXT_ENCODER + '_')[-1].split('_')
                curr_layer = self.text_encoder
            else:
                layer_infos = key.split('.')[0].split(LORA_PREFIX_UNET +
                                                      '_')[-1].split('_')
                curr_layer = self.unet
            # find the target layer
            temp_name = layer_infos.pop(0)
            while len(layer_infos) > -1:
                try:
                    curr_layer = curr_layer.__getattr__(temp_name)
                    if len(layer_infos) > 0:
                        temp_name = layer_infos.pop(0)
                    elif len(layer_infos) == 0:
                        break
                except Exception:
                    if len(temp_name) > 0:
                        temp_name += '_' + layer_infos.pop(0)
                    else:
                        temp_name = layer_infos.pop(0)

            pair_keys = []
            if 'lora_down' in key:
                pair_keys.append(key.replace('lora_down', 'lora_up'))
                pair_keys.append(key)
            else:
                pair_keys.append(key)
                pair_keys.append(key.replace('lora_up', 'lora_down'))

            # update weight
            if len(state_dict[pair_keys[0]].shape) == 4:
                weight_up = state_dict[pair_keys[0]].squeeze(3).squeeze(2).to(
                    torch.float32)
                weight_down = state_dict[pair_keys[1]].squeeze(3).squeeze(
                    2).to(torch.float32)
                curr_layer.weight.data += alpha * torch.mm(
                    weight_up, weight_down).unsqueeze(2).unsqueeze(3).to(
                        curr_layer.weight.data.device)
            else:
                weight_up = state_dict[pair_keys[0]].to(torch.float32)
                weight_down = state_dict[pair_keys[1]].to(torch.float32)
                curr_layer.weight.data += alpha * torch.mm(
                    weight_up, weight_down).to(curr_layer.weight.data.device)

            # update visited list
            for item in pair_keys:
                visited.append(item)

        return self.unet, self.text_encoder

    def prepare_latents(self,
                        batch_size,
                        num_channels_latents,
                        video_length,
                        height,
                        width,
                        dtype,
                        device,
                        generator,
                        latents=None):
        """Prepare latent variables."""
        shape = (batch_size, num_channels_latents, video_length,
                 height // self.vae_scale_factor,
                 width // self.vae_scale_factor)
        if isinstance(generator, list) and len(generator) != batch_size:
            raise ValueError(
                f'You have passed a list of generators of length '
                f'{len(generator)}, but requested an effective batch'
                f' size of {batch_size}. Make sure the batch size matches the'
                f' length of the generators.')
        if latents is None:
            rand_device = 'cpu' if device.type == 'mps' else device

            if isinstance(generator, list):
                shape = shape
                # shape = (1,) + shape[1:]
                latents = [
                    torch.randn(
                        shape,
                        generator=generator[i],
                        device=rand_device,
                        dtype=dtype) for i in range(batch_size)
                ]
                latents = torch.cat(latents, dim=0).to(device)
            else:
                latents = torch.randn(
                    shape,
                    generator=generator,
                    device=rand_device,
                    dtype=dtype).to(device)
        else:
            if latents.shape != shape:
                raise ValueError(f'Unexpected latents shape, got '
                                 f'{latents.shape}, expected {shape}')
            latents = latents.to(device)

        # scale the initial noise by the standard deviation
        # required by the scheduler
        latents = latents * self.scheduler.init_noise_sigma
        return latents

    def prepare_model(self):
        """Prepare model for training.

        Move model to target dtype and disable gradient for some models.
        """
        self.vae.requires_grad_(False)
        print_log('Set VAE untrainable.', 'current')
        self.vae.to(self.dtype)
        print_log(f'Move VAE to {self.dtype}.', 'current')
        # if not self.finetune_text_encoder or self.lora_config:
        if 1:
            self.text_encoder.requires_grad_(False)
            print_log('Set Text Encoder untrainable.', 'current')
            self.text_encoder.to(self.dtype)
            print_log(f'Move Text Encoder to {self.dtype}.', 'current')
        # if self.lora_config:
        if 1:
            self.unet.requires_grad_(False)
            print_log('Set Unet untrainable.', 'current')

    def set_lora(self):
        """Set LORA for model."""
        if self.lora_config:
            set_lora(self.unet, self.lora_config)

    @torch.no_grad()
    def val_step(self, data: dict) -> SampleList:
        """Gets the generated image of given data. Calls
        ``self.data_preprocessor`` and ``self.infer`` in order. Return the
        generated results which will be passed to evaluator or visualizer.

        Args:
            data (dict or tuple or list): Data sampled from dataset.

        Returns:
            SampleList: Generated image or image dict.
        """
        data = self.data_preprocessor(data)
        data_samples = data['data_samples']
        if self.val_prompts is None:
            prompt = data_samples.prompt
        else:
            prompt = self.val_prompts
            # construct a fake data_sample for destruct
            data_samples.split() * len(prompt)
            data_samples = DataSample.stack(data_samples.split() * len(prompt))

        output = self.infer(prompt, return_type='tensor')
        samples = output['samples']
        samples = self.data_preprocessor.destruct(samples, data_samples)

        out_data_sample = DataSample(fake_img=samples, prompt=prompt)
        data_sample_list = out_data_sample.split()
        return data_sample_list

    @torch.no_grad()
    def test_step(self, data: dict) -> SampleList:
        """Gets the generated image of given data. Calls
        ``self.data_preprocessor`` and ``self.infer`` in order. Return the
        generated results which will be passed to evaluator or visualizer.

        Args:
            data (dict or tuple or list): Data sampled from dataset.

        Returns:
            SampleList: Generated image or image dict.
        """
        if self.val_prompts is None:
            data = self.data_preprocessor(data)
            data_samples = data['data_samples']
            prompt = data_samples.prompt
        else:
            prompt = self.val_prompts
            # construct a fake data_sample for destruct
            data_samples = DataSample.stack(data['data_samples'] * len(prompt))

        output = self.infer(prompt, return_type='tensor')
        samples = output['samples']
        samples = self.data_preprocessor.destruct(samples, data_samples)

        out_data_sample = DataSample(fake_img=samples, prompt=prompt)
        data_sample_list = out_data_sample.split()
        return data_sample_list

    @torch.no_grad()
    def infer(self,
              prompt: Union[str, List[str]],
              video_length: Optional[int] = 16,
              height: Optional[int] = None,
              width: Optional[int] = None,
              num_inference_steps: int = 50,
              guidance_scale: float = 7.5,
              negative_prompt: Optional[Union[str, List[str]]] = None,
              num_videos_per_prompt: Optional[int] = 1,
              eta: float = 0.0,
              generator: Optional[Union[torch.Generator,
                                        List[torch.Generator]]] = None,
              latents: Optional[torch.FloatTensor] = None,
              return_type: Optional[str] = 'tensor',
              show_progress: bool = True,
              seed: Optional[int] = 1007):
        """Function invoked when calling the pipeline for generation.

        Args:
            prompt (str or List[str]): The prompt or prompts to guide
                the video generation.
            video_length (int, Option): The number of frames of the
                generated video. Defaults to 16.
            height (int, Optional): The height in pixels of the generated
                image. If not passed, the height will be
                `self.unet_sample_size * self.vae_scale_factor` Defaults
                to None.
            width (int, Optional): The width in pixels of the generated image.
                If not passed, the width will be
                `self.unet_sample_size * self.vae_scale_factor` Defaults
                to None.
            num_inference_steps (int): The number of denoising steps.
                More denoising steps usually lead to a higher quality video at
                the expense of slower inference. Defaults to 50.
            guidance_scale (float): Guidance scale as defined in Classifier-
                Free Diffusion Guidance (https://arxiv.org/abs/2207.12598).
                Defaults to 7.5
            negative_prompt (str or List[str], optional): The prompt or
                prompts not to guide the video generation. Ignored when not
                using guidance (i.e., ignored if `guidance_scale` is less
                than 1). Defaults to None.
            num_videos_per_prompt (int): The number of videos to generate
                per prompt. Defaults to 1.
            eta (float): Corresponds to parameter eta (η) in the DDIM paper:
                https://arxiv.org/abs/2010.02502. Only applies to
                DDIMScheduler, will be ignored for others. Defaults to 0.0.
            generator (torch.Generator, optional): A torch generator to make
                generation deterministic. Defaults to None.
            latents (torch.FloatTensor, optional): Pre-generated noisy latents,
                sampled from a Gaussian distribution, to be used as inputs for
                video generation. Can be used to tweak the same generation with
                different prompts. If not provided, a latents tensor will be
                generated by sampling using the supplied random `generator`.
                Defaults to None.
            return_type (str): The return type of the inference results.
                Supported types are 'video', 'numpy', 'tensor'. If 'video'
                is passed, a list of PIL images will be returned. If 'numpy'
                is passed, a numpy array with shape [N, C, H, W] will be
                returned, and the value range will be same as decoder's
                output range. If 'tensor' is passed, the decoder's output
                will be returned. Defaults to 'image'.
        #TODO
        Returns:
            dict: A dict containing the generated video
        """
        assert return_type in ['image', 'tensor', 'numpy']

        # 0. Default height and width to unet
        height = height or self.unet_sample_size * self.vae_scale_factor
        width = width or self.unet_sample_size * self.vae_scale_factor
        if seed != -1:
            torch.manual_seed(seed)
        print_log(f'current seed: {torch.initial_seed()}')
        print_log(f'sampling {prompt} ...')

        # 1. Check inputs. Raise error if not correct

        self.check_inputs(prompt, height,
                          width)  # NOTE: aligned with origin repo

        # 2. Define call parameters
        batch_size = 1
        if latents is not None:
            batch_size = latents.shape[0]
        if isinstance(prompt, list):
            batch_size = len(prompt)
        device = self.device

        # here `guidance_scale` is defined analog to the
        # guidance weight `w` of equation (2)
        # of the Imagen paper: https://arxiv.org/pdf/2205.11487.pdf .
        # `guidance_scale = 1`
        # corresponds to doing no classifier free guidance.
        do_classifier_free_guidance = guidance_scale > 1.0
        video_dtype = self.vae.module.dtype if hasattr(self.vae, 'module') \
            else self.vae.dtype

        # 3. Encode input prompt

        prompt = prompt if isinstance(prompt, list) else [prompt] * batch_size
        if negative_prompt is not None:
            negative_prompt = negative_prompt if isinstance(
                negative_prompt, list) else [negative_prompt] * batch_size
        text_embeddings = self._encode_prompt(
            prompt, device, num_videos_per_prompt, do_classifier_free_guidance,
            negative_prompt)  # NOTE aligned with origin repo

        # 4. Prepare timesteps
        # self.scheduler.set_timesteps(num_inference_steps, device=device)
        self.test_scheduler.set_timesteps(num_inference_steps)
        timesteps = self.test_scheduler.timesteps

        # 5. Prepare latent variables
        num_channels_latents = self.unet.config.in_channels
        latents = self.prepare_latents(
            batch_size * num_videos_per_prompt,
            num_channels_latents,
            video_length,
            height,
            width,
            text_embeddings.dtype,
            device,
            generator,
            latents,
        )  # NOTE aligned with origin repo
        latents_dtype = latents.dtype

        # 6. Prepare extra step kwargs.
        extra_step_kwargs = self.prepare_extra_step_kwargs(generator, eta)

        # 7. Denoising loop
        if show_progress:
            timesteps = tqdm(timesteps)
        for i, t in enumerate(timesteps):
            # expand the latents if we are doing classifier free guidance
            latent_model_input = torch.cat(
                [latents] * 2) if do_classifier_free_guidance else latents
            latent_model_input = self.test_scheduler.scale_model_input(
                latent_model_input, t)

            # predict the noise residual
            noise_pred = self.unet(
                latent_model_input,
                t,
                encoder_hidden_states=text_embeddings,
            )['sample'].to(dtype=latents_dtype)

            # perform guidance
            if do_classifier_free_guidance:
                noise_pred_uncond, noise_pred_text = noise_pred.chunk(2)
                noise_pred = noise_pred_uncond + guidance_scale * (
                    noise_pred_text - noise_pred_uncond)

                # compute the previous noisy sample x_t -> x_t-1
                latents = self.test_scheduler.step(
                    noise_pred, t, latents,
                    **extra_step_kwargs)['prev_sample']  # FIXME: not aligned
                # FIXME: revise config thresholding=False
                # scheduler pred_original_sample not aligned
                # fixed clip_sample=False
        # 8. Post-processing
        video = self.decode_latents(latents.to(video_dtype))

        if return_type == 'tensor':
            video = torch.from_numpy(video)

        return {'samples': video}

    def forward(self,
                inputs: torch.Tensor,
                data_samples: Optional[list] = None,
                mode: str = 'tensor') -> Union[Dict[str, torch.Tensor], list]:
        """forward is not implemented now."""
        raise NotImplementedError(
            'Forward is not implemented now, please use infer.')



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Tuple

import torch
import torch.nn as nn


class ImgNormalize(nn.Conv2d):
    """Normalize images with the given mean and std value.

    Based on Conv2d layer, can work in GPU.

    Args:
        pixel_range (float): Pixel range of feature.
        img_mean (Tuple[float]): Image mean of each channel.
        img_std (Tuple[float]): Image std of each channel.
        sign (int): Sign of bias. Default -1.
    """

    def __init__(self,
                 pixel_range: float,
                 img_mean: Tuple[float, float, float],
                 img_std: Tuple[float, float, float],
                 sign: int = -1):

        assert len(img_mean) == len(img_std)
        num_channels = len(img_mean)
        super().__init__(num_channels, num_channels, kernel_size=1)

        std = torch.Tensor(img_std)
        self.weight.data = torch.eye(num_channels).view(
            num_channels, num_channels, 1, 1)
        self.weight.data.div_(std.view(num_channels, 1, 1, 1))
        self.bias.data = sign * pixel_range * torch.Tensor(img_mean)
        self.bias.data.div_(std)

        self.weight.requires_grad = False
        self.bias.requires_grad = False



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os.path as osp

import mmengine
import requests

RESOURCES = {
    'Matting': [
        'https://download.openmmlab.com/mmediting/resources/input/matting/GT05.jpg',  # noqa
        'https://download.openmmlab.com/mmediting/resources/input/matting/GT05_trimap.jpg',  # noqa
        'https://download.openmmlab.com/mmediting/resources/input/matting/readme.md'  # noqa
    ],
    'Inpainting': [
        'https://download.openmmlab.com/mmediting/resources/input/inpainting/bbox_mask.png',  # noqa
        'https://download.openmmlab.com/mmediting/resources/input/inpainting/celeba_test.png',  # noqa
        'https://download.openmmlab.com/mmediting/resources/input/inpainting/readme.md'  # noqa
    ],
    'Image Super-Resolution': [
        'https://download.openmmlab.com/mmediting/resources/input/restoration/000001.png',  # noqa
        'https://download.openmmlab.com/mmediting/resources/input/restoration/0901x2.png',  # noqa
        'https://download.openmmlab.com/mmediting/resources/input/restoration/readme.md'  # noqa
    ],
    'Image2Image Translation': [
        'https://download.openmmlab.com/mmediting/resources/input/translation/gt_mask_0.png',  # noqa
        'https://download.openmmlab.com/mmediting/resources/input/translation/readme.md'  # noqa
    ],
    'Video Interpolation': [
        'https://download.openmmlab.com/mmediting/resources/input/video_interpolation/b-3LLDhc4EU_000000_000010.mp4',  # noqa
        'https://download.openmmlab.com/mmediting/resources/input/video_interpolation/readme.md'  # noqa
    ],
    'Video Super-Resolution': [
        'https://download.openmmlab.com/mmediting/resources/input/video_restoration/QUuC4vJs_000084_000094_400x320.mp4',  # noqa
        'https://download.openmmlab.com/mmediting/resources/input/video_restoration/readme.md',  # noqa
        'https://download.openmmlab.com/mmediting/resources/input/video_restoration/v_Basketball_g01_c01.avi'  # noqa
    ]
}


def parse_args():
    parser = argparse.ArgumentParser(description='Download resources')
    parser.add_argument(
        '--root-dir',
        type=str,
        help='resource root dir',
        default='../resources')
    parser.add_argument(
        '--task',
        type=str,
        help='one specific task, if None : download all resources',
        default=None)
    parser.add_argument(
        '--print-all', action='store_true', help='print all resources')
    parser.add_argument(
        '--print-task-type', action='store_true', help='print all task types')
    parser.add_argument(
        '--print-task',
        type=str,
        help='print all tasks that need input resources',
        default=None)

    args = parser.parse_args()
    return args


def main():
    args = parse_args()

    if args.print_all:
        print('all inference resources:')
        for key in RESOURCES.keys():
            print(key)
            for value in RESOURCES[key]:
                print(value)
        return

    if args.print_task_type:
        print('all task type:')
        for key in RESOURCES.keys():
            print(key)
        return

    if args.print_task:
        print('RESOURCES of task ' + args.print_task + ':')
        for value in RESOURCES[args.print_task]:
            print(value)
        return

    to_be_download = []
    if args.task and args.task in RESOURCES.keys():
        to_be_download.extend(RESOURCES[args.task])
    else:
        for key in RESOURCES.keys():
            to_be_download.extend(RESOURCES[key])

    put_root_path = osp.join(osp.dirname(__file__), args.root_dir)
    for item in to_be_download:
        item_relative_path = item[item.find('input'):]
        put_path = osp.join(put_root_path, item_relative_path)
        mmengine.mkdir_or_exist(osp.dirname(put_path))
        response = requests.get(item)
        open(put_path, 'wb').write(response.content)
        print('Download finished: ' + item + ' to ' + put_path)


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
import warnings
from typing import Any, List

from mmagic.utils import try_import
from .ddim_scheduler import EditDDIMScheduler
from .ddpm_scheduler import EditDDPMScheduler


class SchedulerWrapper:
    """Wrapper for schedulers from HuggingFace Diffusers. This wrapper will be
    set a attribute called `_scheduler_cls` by wrapping function and will be
    used to initialize the model structure.

    Example:
    >>> 1. Load pretrained model from HuggingFace Space.
    >>> config = dict(
    >>>     type='DDPMScheduler',
    >>>     from_pretrained='lllyasviel/sd-controlnet-canny',
    >>>     subfolder='scheduler')
    >>> ddpm_scheduler = DIFFUSION_SCHEDULERS.build(config)

    >>> 2. Initialize model with own defined arguments
    >>> config = dict(
    >>>     type='EulerDiscreteScheduler',
    >>>     num_train_timesteps=2000,
    >>>     beta_schedule='scaled_linear')
    >>> euler_scheduler = DIFFUSION_SCHEDULERS.build(config)

    Args:
        from_pretrained (Union[str, os.PathLike], optional): The *model id*
            of a pretrained model or a path to a *directory* containing
            model weights and config. Please refers to
            `diffusers.model.modeling_utils.ModelMixin.from_pretrained`
            for more detail. Defaults to None.

        *args, **kwargs: If `from_pretrained` is passed, *args and **kwargs
            will be passed to `from_pretrained` function. Otherwise, *args
            and **kwargs will be used to initialize the model by
            `self._module_cls(*args, **kwargs)`.
    """

    def __init__(self,
                 from_pretrained=None,
                 from_config=None,
                 *args,
                 **kwargs):

        scheduler_cls = self._scheduler_cls

        self._from_pretrained = from_pretrained
        self._from_config = from_config
        if self._from_pretrained:
            self.scheduler = scheduler_cls.from_pretrained(
                from_pretrained, *args, **kwargs)
        elif self._from_config:
            self.scheduler = scheduler_cls.from_config(from_config, *args,
                                                       **kwargs)
        else:
            self.scheduler = scheduler_cls(*args, **kwargs)

    def __getattr__(self, name: str) -> Any:
        """This function provide a way to access the attributes of the wrapped
        scheduler.

        Args:
            name (str): The name of the attribute.

        Returns:
            Any: The got attribute.
        """

        try:
            return getattr(self.scheduler, name)
        except AttributeError:
            raise AttributeError(f'{name} cannot be found in both '
                                 f'\'{self.__class__.__name__}\' and '
                                 f'\'{self.__class__.__name__}.scheduler\'.')

    def __repr__(self):
        """The representation of the wrapper."""
        s = super().__repr__()
        prefix = f'Wrapped Scheduler Class: {self._scheduler_cls}\n'
        prefix += f'Wrapped Scheduler Name: {self._scheduler_name}\n'
        if self._from_pretrained:
            prefix += f'From Pretrained: {self._from_pretrained}\n'
        if self._from_config:
            prefix += f'From Config: {self._from_config}\n'
        s = prefix + s
        return s


def register_diffusers_schedulers() -> List[str]:
    """Register schedulers in ``diffusers.schedulers`` to the
    ``DIFFUSION_SCHEDULERS`` registry. Specifically, the registered schedulers
    from diffusers define the methodology for iteratively adding noise to an
    image or for updating a sample based on model outputs. See more details
    about schedulers in diffusers here:
    https://huggingface.co/docs/diffusers/api/schedulers/overview.

    Returns:
        List[str]: A list of registered DIFFUSION_SCHEDULERS' name.
    """

    import inspect

    from mmagic.registry import DIFFUSION_SCHEDULERS

    diffusers = try_import('diffusers')
    if diffusers is None:
        warnings.warn('Diffusion Schedulers are not registered as expect. '
                      'If you want to use diffusion models, '
                      'please install diffusers>=0.12.0.')
        return None

    def gen_wrapped_cls(scheduler, scheduler_name):
        return type(
            scheduler_name, (SchedulerWrapper, ),
            dict(
                _scheduler_cls=scheduler,
                _scheduler_name=scheduler_name,
                __module__=__name__))

    DIFFUSERS_SCHEDULERS = []
    for module_name in dir(diffusers.schedulers):
        if module_name.startswith('Flax'):
            continue
        elif module_name.endswith('Scheduler'):
            _scheduler = getattr(diffusers.schedulers, module_name)
            if inspect.isclass(_scheduler):
                wrapped_scheduler = gen_wrapped_cls(_scheduler, module_name)
                DIFFUSION_SCHEDULERS.register_module(
                    name=module_name, module=wrapped_scheduler)
                DIFFUSERS_SCHEDULERS.append(module_name)
    return DIFFUSERS_SCHEDULERS


REGISTERED_DIFFUSERS_SCHEDULERS = register_diffusers_schedulers()

__all__ = ['EditDDIMScheduler', 'EditDDPMScheduler']



_base_ = [
    '../_base_/models/base_liif.py', '../_base_/datasets/liif_test_config.py'
]

experiment_name = 'liif-edsr-norm_c64b16_1xb16-1000k_div2k'
work_dir = f'./work_dirs/{experiment_name}'
save_dir = './work_dirs/'

scale_min, scale_max = 1, 4

# model settings
model = dict(
    type='LIIF',
    generator=dict(
        type='LIIFEDSRNet',
        encoder=dict(
            type='EDSRNet',
            in_channels=3,
            out_channels=3,
            mid_channels=64,
            num_blocks=16),
        imnet=dict(
            type='MLPRefiner',
            in_dim=64,
            out_dim=3,
            hidden_list=[256, 256, 256, 256]),
        local_ensemble=True,
        feat_unfold=True,
        cell_decode=True,
        eval_bsize=30000),
    pixel_loss=dict(type='L1Loss', loss_weight=1.0, reduction='mean'),
    data_preprocessor=dict(
        type='DataPreprocessor',
        mean=[0.4488 * 255, 0.4371 * 255, 0.4040 * 255],
        std=[255., 255., 255.],
    ))



# Copyright (c) OpenMMLab. All rights reserved.
from __future__ import absolute_import, division, print_function
import functools
import math

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils import model_zoo

pretrained_settings = {
    'inceptionresnetv2': {
        'imagenet': {
            'url':
            'https://download.openxlab.org.cn/models/xiaomile/DeblurGANv2/'
            'weight/inceptionresnetv2-520b38e4.pth',
            'input_space': 'RGB',
            'input_size': [3, 299, 299],
            'input_range': [0, 1],
            'mean': [0.5, 0.5, 0.5],
            'std': [0.5, 0.5, 0.5],
            'num_classes': 1000
        },
        'imagenet+background': {
            'url':
            'https://download.openxlab.org.cn/models/xiaomile/DeblurGANv2/'
            'weight/inceptionresnetv2-520b38e4.pth',
            'input_space': 'RGB',
            'input_size': [3, 299, 299],
            'input_range': [0, 1],
            'mean': [0.5, 0.5, 0.5],
            'std': [0.5, 0.5, 0.5],
            'num_classes': 1001
        }
    }
}


class BasicConv2d(nn.Module):

    def __init__(self, in_planes, out_planes, kernel_size, stride, padding=0):
        super(BasicConv2d, self).__init__()
        self.conv = nn.Conv2d(
            in_planes,
            out_planes,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            bias=False)  # verify bias false
        self.bn = nn.BatchNorm2d(
            out_planes,
            eps=0.001,  # value found in tensorflow
            momentum=0.1,  # default pytorch value
            affine=True)
        self.relu = nn.ReLU(inplace=False)

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


class Mixed_5b(nn.Module):

    def __init__(self):
        super(Mixed_5b, self).__init__()

        self.branch0 = BasicConv2d(192, 96, kernel_size=1, stride=1)

        self.branch1 = nn.Sequential(
            BasicConv2d(192, 48, kernel_size=1, stride=1),
            BasicConv2d(48, 64, kernel_size=5, stride=1, padding=2))

        self.branch2 = nn.Sequential(
            BasicConv2d(192, 64, kernel_size=1, stride=1),
            BasicConv2d(64, 96, kernel_size=3, stride=1, padding=1),
            BasicConv2d(96, 96, kernel_size=3, stride=1, padding=1))

        self.branch3 = nn.Sequential(
            nn.AvgPool2d(3, stride=1, padding=1, count_include_pad=False),
            BasicConv2d(192, 64, kernel_size=1, stride=1))

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        x3 = self.branch3(x)
        out = torch.cat((x0, x1, x2, x3), 1)
        return out


class Block35(nn.Module):

    def __init__(self, scale=1.0):
        super(Block35, self).__init__()

        self.scale = scale

        self.branch0 = BasicConv2d(320, 32, kernel_size=1, stride=1)

        self.branch1 = nn.Sequential(
            BasicConv2d(320, 32, kernel_size=1, stride=1),
            BasicConv2d(32, 32, kernel_size=3, stride=1, padding=1))

        self.branch2 = nn.Sequential(
            BasicConv2d(320, 32, kernel_size=1, stride=1),
            BasicConv2d(32, 48, kernel_size=3, stride=1, padding=1),
            BasicConv2d(48, 64, kernel_size=3, stride=1, padding=1))

        self.conv2d = nn.Conv2d(128, 320, kernel_size=1, stride=1)
        self.relu = nn.ReLU(inplace=False)

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        out = torch.cat((x0, x1, x2), 1)
        out = self.conv2d(out)
        out = out * self.scale + x
        out = self.relu(out)
        return out


class Mixed_6a(nn.Module):

    def __init__(self):
        super(Mixed_6a, self).__init__()

        self.branch0 = BasicConv2d(320, 384, kernel_size=3, stride=2)

        self.branch1 = nn.Sequential(
            BasicConv2d(320, 256, kernel_size=1, stride=1),
            BasicConv2d(256, 256, kernel_size=3, stride=1, padding=1),
            BasicConv2d(256, 384, kernel_size=3, stride=2))

        self.branch2 = nn.MaxPool2d(3, stride=2)

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        out = torch.cat((x0, x1, x2), 1)
        return out


class Block17(nn.Module):

    def __init__(self, scale=1.0):
        super(Block17, self).__init__()

        self.scale = scale

        self.branch0 = BasicConv2d(1088, 192, kernel_size=1, stride=1)

        self.branch1 = nn.Sequential(
            BasicConv2d(1088, 128, kernel_size=1, stride=1),
            BasicConv2d(
                128, 160, kernel_size=(1, 7), stride=1, padding=(0, 3)),
            BasicConv2d(
                160, 192, kernel_size=(7, 1), stride=1, padding=(3, 0)))

        self.conv2d = nn.Conv2d(384, 1088, kernel_size=1, stride=1)
        self.relu = nn.ReLU(inplace=False)

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        out = torch.cat((x0, x1), 1)
        out = self.conv2d(out)
        out = out * self.scale + x
        out = self.relu(out)
        return out


class Mixed_7a(nn.Module):

    def __init__(self):
        super(Mixed_7a, self).__init__()

        self.branch0 = nn.Sequential(
            BasicConv2d(1088, 256, kernel_size=1, stride=1),
            BasicConv2d(256, 384, kernel_size=3, stride=2))

        self.branch1 = nn.Sequential(
            BasicConv2d(1088, 256, kernel_size=1, stride=1),
            BasicConv2d(256, 288, kernel_size=3, stride=2))

        self.branch2 = nn.Sequential(
            BasicConv2d(1088, 256, kernel_size=1, stride=1),
            BasicConv2d(256, 288, kernel_size=3, stride=1, padding=1),
            BasicConv2d(288, 320, kernel_size=3, stride=2))

        self.branch3 = nn.MaxPool2d(3, stride=2)

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        x2 = self.branch2(x)
        x3 = self.branch3(x)
        out = torch.cat((x0, x1, x2, x3), 1)
        return out


class Block8(nn.Module):

    def __init__(self, scale=1.0, noReLU=False):
        super(Block8, self).__init__()

        self.scale = scale
        self.noReLU = noReLU

        self.branch0 = BasicConv2d(2080, 192, kernel_size=1, stride=1)

        self.branch1 = nn.Sequential(
            BasicConv2d(2080, 192, kernel_size=1, stride=1),
            BasicConv2d(
                192, 224, kernel_size=(1, 3), stride=1, padding=(0, 1)),
            BasicConv2d(
                224, 256, kernel_size=(3, 1), stride=1, padding=(1, 0)))

        self.conv2d = nn.Conv2d(448, 2080, kernel_size=1, stride=1)
        if not self.noReLU:
            self.relu = nn.ReLU(inplace=False)

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        x0 = self.branch0(x)
        x1 = self.branch1(x)
        out = torch.cat((x0, x1), 1)
        out = self.conv2d(out)
        out = out * self.scale + x
        if not self.noReLU:
            out = self.relu(out)
        return out


class InceptionResNetV2(nn.Module):
    """Define a inceptionresnetv2 model."""

    def __init__(self, num_classes=1001):
        super(InceptionResNetV2, self).__init__()
        # Special attributes
        self.input_space = None
        self.input_size = (299, 299, 3)
        self.mean = None
        self.std = None
        # Modules
        self.conv2d_1a = BasicConv2d(3, 32, kernel_size=3, stride=2)
        self.conv2d_2a = BasicConv2d(32, 32, kernel_size=3, stride=1)
        self.conv2d_2b = BasicConv2d(
            32, 64, kernel_size=3, stride=1, padding=1)
        self.maxpool_3a = nn.MaxPool2d(3, stride=2)
        self.conv2d_3b = BasicConv2d(64, 80, kernel_size=1, stride=1)
        self.conv2d_4a = BasicConv2d(80, 192, kernel_size=3, stride=1)
        self.maxpool_5a = nn.MaxPool2d(3, stride=2)
        self.mixed_5b = Mixed_5b()
        self.repeat = nn.Sequential(
            Block35(scale=0.17), Block35(scale=0.17), Block35(scale=0.17),
            Block35(scale=0.17), Block35(scale=0.17), Block35(scale=0.17),
            Block35(scale=0.17), Block35(scale=0.17), Block35(scale=0.17),
            Block35(scale=0.17))
        self.mixed_6a = Mixed_6a()
        self.repeat_1 = nn.Sequential(
            Block17(scale=0.10), Block17(scale=0.10), Block17(scale=0.10),
            Block17(scale=0.10), Block17(scale=0.10), Block17(scale=0.10),
            Block17(scale=0.10), Block17(scale=0.10), Block17(scale=0.10),
            Block17(scale=0.10), Block17(scale=0.10), Block17(scale=0.10),
            Block17(scale=0.10), Block17(scale=0.10), Block17(scale=0.10),
            Block17(scale=0.10), Block17(scale=0.10), Block17(scale=0.10),
            Block17(scale=0.10), Block17(scale=0.10))
        self.mixed_7a = Mixed_7a()
        self.repeat_2 = nn.Sequential(
            Block8(scale=0.20), Block8(scale=0.20), Block8(scale=0.20),
            Block8(scale=0.20), Block8(scale=0.20), Block8(scale=0.20),
            Block8(scale=0.20), Block8(scale=0.20), Block8(scale=0.20))
        self.block8 = Block8(noReLU=True)
        self.conv2d_7b = BasicConv2d(2080, 1536, kernel_size=1, stride=1)
        self.avgpool_1a = nn.AvgPool2d(8, count_include_pad=False)
        self.last_linear = nn.Linear(1536, num_classes)

    def features(self, input):
        """Get network features.

        Args:
            input (torch.tensor): You can directly input a ``torch.Tensor``.
        """
        x = self.conv2d_1a(input)
        x = self.conv2d_2a(x)
        x = self.conv2d_2b(x)
        x = self.maxpool_3a(x)
        x = self.conv2d_3b(x)
        x = self.conv2d_4a(x)
        x = self.maxpool_5a(x)
        x = self.mixed_5b(x)
        x = self.repeat(x)
        x = self.mixed_6a(x)
        x = self.repeat_1(x)
        x = self.mixed_7a(x)
        x = self.repeat_2(x)
        x = self.block8(x)
        x = self.conv2d_7b(x)
        return x

    def logits(self, features):
        """Get features logits.

        Args:
            features (torch.tensor): You can directly input a ``torch.Tensor``.
        """
        x = self.avgpool_1a(features)
        x = x.view(x.size(0), -1)
        x = F.dropout(x, training=self.training)
        x = self.last_linear(x)
        return x

    def forward(self, input):
        """Forward function.

        Args:
            input (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        x = self.features(input)
        x = self.logits(x)
        return x


def inceptionresnetv2(num_classes=1000, pretrained='imagenet'):
    """return a inceptionresnetv2 network."""
    if pretrained:
        settings = pretrained_settings['inceptionresnetv2'][pretrained]
        model = InceptionResNetV2(num_classes=1001)
        model.load_state_dict(model_zoo.load_url(settings['url']))

        if pretrained == 'imagenet':
            new_last_linear = nn.Linear(1536, 1000)
            new_last_linear.weight.data = model.last_linear.weight.data[1:]
            new_last_linear.bias.data = model.last_linear.bias.data[1:]
            model.last_linear = new_last_linear

        model.input_space = settings['input_space']
        model.input_size = settings['input_size']
        model.input_range = settings['input_range']

        model.mean = settings['mean']
        model.std = settings['std']
    else:
        model = InceptionResNetV2(num_classes=num_classes)
    return model


def conv_bn(inp, oup, stride):
    return nn.Sequential(
        nn.Conv2d(inp, oup, 3, stride, 1, bias=False), nn.BatchNorm2d(oup),
        nn.ReLU6(inplace=True))


def conv_1x1_bn(inp, oup):
    return nn.Sequential(
        nn.Conv2d(inp, oup, 1, 1, 0, bias=False), nn.BatchNorm2d(oup),
        nn.ReLU6(inplace=True))


class InvertedResidual(nn.Module):

    def __init__(self, inp, oup, stride, expand_ratio):
        super(InvertedResidual, self).__init__()
        self.stride = stride
        assert stride in [1, 2]

        hidden_dim = round(inp * expand_ratio)
        self.use_res_connect = self.stride == 1 and inp == oup

        if expand_ratio == 1:
            self.conv = nn.Sequential(
                # dw
                nn.Conv2d(
                    hidden_dim,
                    hidden_dim,
                    3,
                    stride,
                    1,
                    groups=hidden_dim,
                    bias=False),
                nn.BatchNorm2d(hidden_dim),
                nn.ReLU6(inplace=True),
                # pw-linear
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
                nn.BatchNorm2d(oup),
            )
        else:
            self.conv = nn.Sequential(
                # pw
                nn.Conv2d(inp, hidden_dim, 1, 1, 0, bias=False),
                nn.BatchNorm2d(hidden_dim),
                nn.ReLU6(inplace=True),
                # dw
                nn.Conv2d(
                    hidden_dim,
                    hidden_dim,
                    3,
                    stride,
                    1,
                    groups=hidden_dim,
                    bias=False),
                nn.BatchNorm2d(hidden_dim),
                nn.ReLU6(inplace=True),
                # pw-linear
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
                nn.BatchNorm2d(oup),
            )

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)


class MobileNetV2(nn.Module):

    def __init__(self, n_class=1000, input_size=224, width_mult=1.):
        super(MobileNetV2, self).__init__()
        block = InvertedResidual
        input_channel = 32
        last_channel = 1280
        interverted_residual_setting = [
            # t, c, n, s
            [1, 16, 1, 1],
            [6, 24, 2, 2],
            [6, 32, 3, 2],
            [6, 64, 4, 2],
            [6, 96, 3, 1],
            [6, 160, 3, 2],
            [6, 320, 1, 1],
        ]

        # building first layer
        assert input_size % 32 == 0
        input_channel = int(input_channel * width_mult)
        self.last_channel = int(
            last_channel * width_mult) if width_mult > 1.0 else last_channel
        self.features = [conv_bn(3, input_channel, 2)]
        # building inverted residual blocks
        for t, c, n, s in interverted_residual_setting:
            output_channel = int(c * width_mult)
            for i in range(n):
                if i == 0:
                    self.features.append(
                        block(
                            input_channel, output_channel, s, expand_ratio=t))
                else:
                    self.features.append(
                        block(
                            input_channel, output_channel, 1, expand_ratio=t))
                input_channel = output_channel
        # building last several layers
        self.features.append(conv_1x1_bn(input_channel, self.last_channel))
        # make it nn.Sequential
        self.features = nn.Sequential(*self.features)

        # building classifier
        self.classifier = nn.Sequential(
            nn.Dropout(0.2),
            nn.Linear(self.last_channel, n_class),
        )

        self._initialize_weights()

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        x = self.features(x)
        x = x.mean(3).mean(2)
        x = self.classifier(x)
        return x

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                n = m.weight.size(1)
                m.weight.data.normal_(0, 0.01)
                m.bias.data.zero_()


def get_norm_layer(norm_type='instance'):
    """Returns a norm layer of the specified type.

    Args:
        norm_type (Str): norm layer type
    """
    if norm_type == 'batch':
        norm_layer = functools.partial(nn.BatchNorm2d, affine=True)
    elif norm_type == 'instance':
        norm_layer = functools.partial(
            nn.InstanceNorm2d, affine=False, track_running_stats=True)
    else:
        raise NotImplementedError('normalization layer [%s] is not found' %
                                  norm_type)
    return norm_layer



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Callable, Dict, List, Sequence, Tuple, Union

from mmengine.config import ConfigDict
from mmengine.structures import BaseDataElement
from torch import Tensor

ForwardInputs = Tuple[Dict[str, Union[Tensor, str, int]], Tensor]
SampleList = Sequence[BaseDataElement]

NoiseVar = Union[Tensor, Callable, None]
LabelVar = Union[Tensor, Callable, List[int], None]

ConfigType = Union[ConfigDict, Dict]



_base_ = ['swinir_psnr-x4s64w8d6e180_8xb4-lr1e-4-600k_df2k-ost.py']

experiment_name = 'swinir_psnr-x4s64w8d9e240_8xb4-lr1e-4-600k_df2k-ost'
work_dir = f'./work_dirs/{experiment_name}'
save_dir = './work_dirs/'

# model settings
model = dict(
    generator=dict(
        depths=[6, 6, 6, 6, 6, 6, 6, 6, 6],
        embed_dim=240,
        num_heads=[8, 8, 8, 8, 8, 8, 8, 8, 8],
        resi_connection='3conv'))



# Copyright (c) OpenMMLab. All rights reserved.
import collections.abc
from itertools import repeat


# From PyTorch internals
def _ntuple(n):
    """A `to_tuple` function generator. It returns a function, this function
    will repeat the input to a tuple of length ``n`` if the input is not an
    Iterable object, otherwise, return the input directly.

    Args:
        n (int): The number of the target length.
    """

    def parse(x):
        if isinstance(x, collections.abc.Iterable) and not isinstance(x, str):
            return x
        return tuple(repeat(x, n))

    return parse


to_2tuple = _ntuple(2)


def drop_path(x,
              drop_prob: float = 0.,
              training: bool = False,
              scale_by_keep: bool = True):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of
    residual blocks).

    This is the same as the DropConnect impl I created for
    EfficientNet, etc networks, however, the original name is misleading
    as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion:
    https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956
    I've opted for changing the layer and argument names to 'drop path'
    rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.
    """
    if drop_prob == 0. or not training:
        return x
    keep_prob = 1 - drop_prob
    # work with diff dim tensors, not just 2D ConvNets
    shape = (x.shape[0], ) + (1, ) * (x.ndim - 1)
    random_tensor = x.new_empty(shape).bernoulli_(keep_prob)
    if keep_prob > 0.0 and scale_by_keep:
        random_tensor.div_(keep_prob)
    return x * random_tensor


def window_partition(x, window_size):
    """
    Args:
        x: (B, H, W, C)
        window_size (int): window size
    Returns:
        windows: (num_windows*B, window_size, window_size, C)
    """
    B, H, W, C = x.shape
    x = x.view(B, H // window_size, window_size, W // window_size, window_size,
               C)
    windows = x.permute(0, 1, 3, 2, 4,
                        5).contiguous().view(-1, window_size, window_size, C)
    return windows


def window_reverse(windows, window_size, H, W):
    """
    Args:
        windows: (num_windows*B, window_size, window_size, C)
        window_size (int): Window size
        H (int): Height of image
        W (int): Width of image
    Returns:
        x: (B, H, W, C)
    """
    B = int(windows.shape[0] / (H * W / window_size / window_size))
    x = windows.view(B, H // window_size, W // window_size, window_size,
                     window_size, -1)
    x = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(B, H, W, -1)
    return x



# Copyright (c) OpenMMLab. All rights reserved.
import gc

import gradio as gr
import numpy as np
import PIL
import torch
from mmengine import Config

from mmagic.registry import MODELS
from mmagic.utils import register_all_modules

gc.collect()
torch.cuda.empty_cache()
register_all_modules()


class ModelWrapper:

    def __init__(self, model):
        super().__init__()
        self.model = model
        self.device = torch.device(
            'cuda' if torch.cuda.is_available() else 'cpu')
        self.model = self.model.to(self.device)

    def inference(
        self,
        image1: PIL.Image.Image,
        image2: PIL.Image.Image,
        prompt: str,
        negative_prompt: str,
        seed: int,
        guidance_scale: float,
        alpha_: float,
        num_steps: int,
        num_images: int,
    ):
        print('Running model inference...')
        image = []
        if image1 is not None:
            image.append(image1)

        if image2 is not None:
            image.append(image2)

        if len(image) == 0:
            return [], 'You need to upload at least one image.'

        num_subject_in_text = (np.array(
            self.model.special_tokenizer.encode(prompt)) ==
                               self.model.image_token_id).sum()
        if num_subject_in_text != len(image):
            return (
                [],
                "Number of subjects in the text description doesn't "
                'match the number of reference images, #text subjects: '
                f'{num_subject_in_text} #reference image: {len(image)}',
            )

        if seed == -1:
            seed = np.random.randint(0, 1000000)

        generator = torch.Generator(device=self.device)
        generator.manual_seed(seed)

        return (
            self.model.infer(
                prompt=prompt,
                negative_prompt=negative_prompt,
                height=512,
                width=512,
                num_inference_steps=num_steps,
                guidance_scale=guidance_scale,
                num_images_per_prompt=num_images,
                generator=generator,
                alpha_=alpha_,
                reference_subject_images=image,
            )['samples'],
            'run successfully',
        )


def create_demo():
    TITLE = 'FastComposer Demo'

    DESCRIPTION = """To run the demo, you should:
    1. Upload your images. The order of image1 and image2 needs to match the
    order of the subects in the prompt. You only need 1 image for
    single subject generation.
    2. Input proper text prompts, such as "A woman img and a man img in
    the snow" or "A painting of a man img in the style of Van Gogh",
    where "img" specifies the token you want to augment and comes
    after the word.
    3. Click the Run button. You can also adjust the hyperparameters to
    improve the results. Look at the job status to see
    if there are any errors with your input.
    """

    cfg_file = Config.fromfile(
        'configs/fastcomposer/fastcomposer_8xb16_FFHQ.py')
    fastcomposer = MODELS.build(cfg_file.model)
    model = ModelWrapper(fastcomposer)
    # model = ModelWrapper(convert_model_to_pipeline(args, "cpu"))
    with gr.Blocks() as demo:
        gr.Markdown(TITLE)
        gr.Markdown(DESCRIPTION)
        with gr.Row():
            with gr.Column():
                with gr.Box():
                    image1 = gr.Image(label='Image 1', type='pil')

                    image2 = gr.Image(label='Image 2', type='pil')

                    gr.Markdown('Upload the image for your subject')

                prompt = gr.Text(
                    value='A man img and a man img sitting in a park',
                    label='Prompt',
                    placeholder=
                    'e.g. "A woman img and a man img in the snow", "A painting'
                    ' of a man img in the style of Van Gogh"',
                    info='Use "img" to specify the word you want to augment.',
                )
                negative_prompt = gr.Text(
                    value='((((ugly)))), (((duplicate))), ((morbid)), '
                    '((mutilated)), [out of frame], extra fingers, '
                    'mutated hands, ((poorly drawn hands)), '
                    '((poorly drawn face)), (((mutation))), '
                    '(((deformed))), ((ugly)), blurry, ((bad anatomy)), '
                    '(((bad proportions))), ((extra limbs)), cloned face, '
                    '(((disfigured))). out of frame, ugly, extra limbs, '
                    '(bad anatomy), gross proportions, (malformed limbs), '
                    '((missing arms)), ((missing legs)), (((extra arms))), '
                    '(((extra legs))), mutated hands, (fused fingers), '
                    '(too many fingers), (((long neck)))',
                    label='Negative Prompt',
                    info='Features that you want to avoid.',
                )
                alpha_ = gr.Slider(
                    label='alpha',
                    minimum=0,
                    maximum=1,
                    step=0.05,
                    value=0.75,
                    info=
                    'A smaller alpha aligns images with text better, but may '
                    'deviate from the subject image. Increase alpha to improve'
                    ' identity preservation, decrease it for '
                    'prompt consistency.',
                )
                num_images = gr.Slider(
                    label='Number of generated images',
                    minimum=1,
                    maximum=8,
                    step=1,
                    value=4,
                )
                run_button = gr.Button('Run')
                with gr.Accordion(label='Advanced options', open=False):
                    seed = gr.Slider(
                        label='Seed',
                        minimum=-1,
                        maximum=1000000,
                        step=1,
                        value=-1,
                        info='If set to -1, a different seed will be '
                        'used each time.',
                    )
                    guidance_scale = gr.Slider(
                        label='Guidance scale',
                        minimum=1,
                        maximum=10,
                        step=1,
                        value=5,
                    )
                    num_steps = gr.Slider(
                        label='Steps',
                        minimum=1,
                        maximum=300,
                        step=1,
                        value=50,
                    )
            with gr.Column():
                result = gr.Gallery(label='Generated Images').style(
                    grid=[2], height='auto')
                error_message = gr.Text(label='Job Status')

        inputs = [
            image1,
            image2,
            prompt,
            negative_prompt,
            seed,
            guidance_scale,
            alpha_,
            num_steps,
            num_images,
        ]
        run_button.click(
            fn=model.inference, inputs=inputs, outputs=[result, error_message])
    return demo


if __name__ == '__main__':
    demo = create_demo()
    demo.queue(api_open=False)
    demo.launch(
        show_error=True,
        server_name='0.0.0.0',
        server_port=8080,
    )



# Copyright (c) OpenMMLab. All rights reserved.
import os

import gradio as gr
import torch
from mmengine import Config
from mmengine.runner import Runner
from PIL import Image

from mmagic.registry import MODELS
from mmagic.utils import register_all_modules

register_all_modules()

checkpoint_dir = 'ckpts'
ckpts = os.listdir(checkpoint_dir)


def load_base_model():
    global cfg, model
    cfg = Config.fromfile('configs/vico/vico.py')
    model = MODELS.build(cfg.model)


def train(train_data, init_token, placeholder):
    data_root = 'tmp'
    concept_dir = init_token
    image_dir = os.path.join(data_root, concept_dir)
    os.makedirs(image_dir, exist_ok=True)
    for i, data in enumerate(train_data):
        image = Image.open(data.name)
        image.save(os.path.join(image_dir, f'{i}.png'))
    global train_cfg
    train_cfg = Config.fromfile('configs/vico/vico.py')
    train_cfg.work_dir = os.path.join('./work_dirs', 'vico_gradio')
    train_cfg.dataset['data_root'] = data_root
    train_cfg.dataset['concept_dir'] = init_token
    train_cfg.dataset['placeholder'] = placeholder
    train_cfg.train_dataloader['dataset'] = train_cfg.dataset
    train_cfg.model['placeholder'] = placeholder
    train_cfg.model['initialize_token'] = init_token
    train_cfg.custom_hooks = None
    runner = Runner.from_cfg(train_cfg)
    runner.train()


class DummyModel:

    def __init__(self, model):
        self.model = model


def infer_fn(checkpoint, img_ref, prompt, negative_prompt, guidance_scale,
             width, height, seed, inference_steps, bs):
    state_dict = torch.load(os.path.join(checkpoint_dir, checkpoint))
    dummy_agent = DummyModel(model)
    dummy_agent.model.load_state_dict(state_dict, strict=False)
    dummy_agent.model = dummy_agent.model.cuda()
    img_ref = Image.open(img_ref)
    with torch.no_grad():
        output = dummy_agent.model.infer(
            prompt=prompt,
            image_reference=img_ref,
            width=width,
            height=height,
            negative_prompt=negative_prompt,
            guidance_scale=guidance_scale,
            num_inference_steps=inference_steps,
            num_images_per_prompt=bs,
            seed=int(seed))['samples']

    return output


if __name__ == '__main__':
    load_base_model()
    block = gr.Blocks().queue()
    with block:
        with gr.Tab('Train'):
            with gr.Row():
                gr.Markdown('## ViCo')
            with gr.Row():
                with gr.Column():
                    train_data = gr.File(
                        label='Training Samples',
                        file_count='directory',
                        file_types=['image'],
                        interactive=True,
                    )
                with gr.Column():
                    init_token = gr.Textbox(label='Init token')
                    placeholder = gr.Textbox(label='Placeholder')
                    train_button = gr.Button(value='Start Training')
        train_button.click(
            fn=train, inputs=[train_data, init_token, placeholder])
        with gr.Tab('Inference'):
            with gr.Row():
                gr.Markdown('## ViCo')
            with gr.Row():
                with gr.Column():
                    checkpoint = gr.Dropdown(ckpts)
                    img_ref = gr.Image(
                        label='Image Reference',
                        source='upload',
                        type='filepath',
                        elem_id='input-vid')
                    prompt = gr.Textbox(label='Prompt')
                    with gr.Accordion('Advanced options', open=False):
                        n_prompt = gr.Textbox(
                            label='Negative Prompt', value='')
                        guidance_scale = gr.Slider(
                            label='CFG',
                            minimum=0.0,
                            maximum=13.0,
                            value=7.5,
                            step=0.5)
                        width = gr.Slider(
                            label='Image Width',
                            minimum=256,
                            maximum=768,
                            value=512,
                            step=64)
                        height = gr.Slider(
                            label='Image Width',
                            minimum=256,
                            maximum=768,
                            value=512,
                            step=64)
                        seed = gr.Number(
                            label='Seed',
                            value=0,
                        )
                        num_inference_steps = gr.Slider(
                            label='Inference Steps',
                            minimum=20,
                            maximum=80,
                            value=50,
                            step=5)
                        bs = gr.Slider(
                            label='Batch Size',
                            minimum=1,
                            maximum=4,
                            value=1,
                            step=1)

                with gr.Column():
                    output = gr.Gallery(
                        label='Output Image', elem_id='image-output')
                    run_button = gr.Button(label='Run')

            ips = [
                checkpoint, img_ref, prompt, n_prompt, guidance_scale, width,
                height, seed, num_inference_steps, bs
            ]
            run_button.click(fn=infer_fn, inputs=ips, outputs=output)
    block.launch(server_name='0.0.0.0', share=True)



# Copyright (c) OpenMMLab. All rights reserved.
import functools

import numpy as np
import torch.nn as nn

from mmagic.registry import MODELS
from .deblurganv2_util import get_norm_layer

backbone_list = ['DoubleGan', 'MultiScale', 'NoGan', 'PatchGan']


class NLayerDiscriminator(nn.Module):
    """Defines the PatchGAN discriminator with the specified arguments."""

    def __init__(self,
                 input_nc=3,
                 ndf=64,
                 n_layers=3,
                 norm_layer=nn.BatchNorm2d,
                 use_sigmoid=False,
                 use_parallel=True):
        super(NLayerDiscriminator, self).__init__()
        self.use_parallel = use_parallel
        if type(norm_layer) == functools.partial:
            use_bias = norm_layer.func == nn.InstanceNorm2d
        else:
            use_bias = norm_layer == nn.InstanceNorm2d

        kw = 4
        padw = int(np.ceil((kw - 1) / 2))
        sequence = [
            nn.Conv2d(input_nc, ndf, kernel_size=kw, stride=2, padding=padw),
            nn.LeakyReLU(0.2, True)
        ]

        nf_mult = 1
        for n in range(1, n_layers):
            nf_mult_prev = nf_mult
            nf_mult = min(2**n, 8)
            sequence += [
                nn.Conv2d(
                    ndf * nf_mult_prev,
                    ndf * nf_mult,
                    kernel_size=kw,
                    stride=2,
                    padding=padw,
                    bias=use_bias),
                norm_layer(ndf * nf_mult),
                nn.LeakyReLU(0.2, True)
            ]

        nf_mult_prev = nf_mult
        nf_mult = min(2**n_layers, 8)
        sequence += [
            nn.Conv2d(
                ndf * nf_mult_prev,
                ndf * nf_mult,
                kernel_size=kw,
                stride=1,
                padding=padw,
                bias=use_bias),
            norm_layer(ndf * nf_mult),
            nn.LeakyReLU(0.2, True)
        ]

        sequence += [
            nn.Conv2d(
                ndf * nf_mult, 1, kernel_size=kw, stride=1, padding=padw)
        ]

        if use_sigmoid:
            sequence += [nn.Sigmoid()]

        self.model = nn.Sequential(*sequence)

    def forward(self, input):
        """Forward function.

        Args:
            input (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        return self.model(input)


class DicsriminatorTail(nn.Module):

    def __init__(self,
                 nf_mult,
                 n_layers,
                 ndf=64,
                 norm_layer=nn.BatchNorm2d,
                 use_parallel=True):
        super(DicsriminatorTail, self).__init__()
        self.use_parallel = use_parallel
        if type(norm_layer) == functools.partial:
            use_bias = norm_layer.func == nn.InstanceNorm2d
        else:
            use_bias = norm_layer == nn.InstanceNorm2d

        kw = 4
        padw = int(np.ceil((kw - 1) / 2))

        nf_mult_prev = nf_mult
        nf_mult = min(2**n_layers, 8)
        sequence = [
            nn.Conv2d(
                ndf * nf_mult_prev,
                ndf * nf_mult,
                kernel_size=kw,
                stride=1,
                padding=padw,
                bias=use_bias),
            norm_layer(ndf * nf_mult),
            nn.LeakyReLU(0.2, True)
        ]

        sequence += [
            nn.Conv2d(
                ndf * nf_mult, 1, kernel_size=kw, stride=1, padding=padw)
        ]

        self.model = nn.Sequential(*sequence)

    def forward(self, input):
        """Forward function.

        Args:
            input (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        return self.model(input)


class MultiScaleDiscriminator(nn.Module):
    """Defines the MultiScale PatchGAN discriminator with the specified
    arguments."""

    def __init__(self,
                 input_nc=3,
                 ndf=64,
                 norm_layer=nn.BatchNorm2d,
                 use_parallel=True):
        super(MultiScaleDiscriminator, self).__init__()
        self.use_parallel = use_parallel
        if type(norm_layer) == functools.partial:
            use_bias = norm_layer.func == nn.InstanceNorm2d
        else:
            use_bias = norm_layer == nn.InstanceNorm2d

        kw = 4
        padw = int(np.ceil((kw - 1) / 2))
        sequence = [
            nn.Conv2d(input_nc, ndf, kernel_size=kw, stride=2, padding=padw),
            nn.LeakyReLU(0.2, True)
        ]

        nf_mult = 1
        for n in range(1, 3):
            nf_mult_prev = nf_mult
            nf_mult = min(2**n, 8)
            sequence += [
                nn.Conv2d(
                    ndf * nf_mult_prev,
                    ndf * nf_mult,
                    kernel_size=kw,
                    stride=2,
                    padding=padw,
                    bias=use_bias),
                norm_layer(ndf * nf_mult),
                nn.LeakyReLU(0.2, True)
            ]

        self.scale_one = nn.Sequential(*sequence)
        self.first_tail = DicsriminatorTail(nf_mult=nf_mult, n_layers=3)
        nf_mult_prev = 4
        nf_mult = 8

        self.scale_two = nn.Sequential(
            nn.Conv2d(
                ndf * nf_mult_prev,
                ndf * nf_mult,
                kernel_size=kw,
                stride=2,
                padding=padw,
                bias=use_bias), norm_layer(ndf * nf_mult),
            nn.LeakyReLU(0.2, True))
        nf_mult_prev = nf_mult
        self.second_tail = DicsriminatorTail(nf_mult=nf_mult, n_layers=4)
        self.scale_three = nn.Sequential(
            nn.Conv2d(
                ndf * nf_mult_prev,
                ndf * nf_mult,
                kernel_size=kw,
                stride=2,
                padding=padw,
                bias=use_bias), norm_layer(ndf * nf_mult),
            nn.LeakyReLU(0.2, True))
        self.third_tail = DicsriminatorTail(nf_mult=nf_mult, n_layers=5)

    def forward(self, input):
        """Forward function.

        Args:
            input (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        x = self.scale_one(input)
        x_1 = self.first_tail(x)
        x = self.scale_two(x)
        x_2 = self.second_tail(x)
        x = self.scale_three(x)
        x = self.third_tail(x)
        return [x_1, x_2, x]


def get_fullD(norm_layer):
    """Get a full gan discriminator.

    Args:
        norm_layer (Str): norm type
    """
    model_d = NLayerDiscriminator(
        n_layers=5,
        norm_layer=get_norm_layer(norm_type=norm_layer),
        use_sigmoid=False)
    return model_d


class DoubleGan(nn.Module):
    """Get a discriminator with a patch gan and a full gan."""

    def __init__(self, norm_layer='instance', d_layers=3):
        super().__init__()
        self.patch_gan = NLayerDiscriminator(
            n_layers=d_layers,
            norm_layer=get_norm_layer(norm_type=norm_layer),
            use_sigmoid=False)
        self.full_gan = get_fullD(norm_layer)

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            List(torch.Tensor) : ``List(torch.tensor)`` will be returned.
        """
        # d_full_gan = self.model_d['full'](x)
        d_full_gan_output = self.full_gan(x)
        # d_patch_gan = self.model_d['patch'](x)
        d_patch_gan_output = self.patch_gan(x)
        return [d_full_gan_output, d_patch_gan_output]


class PatchGan(nn.Module):
    """A patch gan discriminator with the specified arguments."""

    def __init__(self, norm_layer='instance', d_layers=3):
        super().__init__()
        self.patch_gan = NLayerDiscriminator(
            n_layers=d_layers,
            norm_layer=get_norm_layer(norm_type=norm_layer),
            use_sigmoid=False)

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        d_patch_gan_output = self.patch_gan(x)
        return d_patch_gan_output


class MultiScale(nn.Module):
    """A multiscale patch gan discriminator with the specified arguments."""

    def __init__(self, norm_layer='instance', d_layers=3):
        super().__init__()
        self.model_d = MultiScaleDiscriminator(
            norm_layer=get_norm_layer(norm_type=norm_layer))

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor ): You can directly input a ``torch.Tensor``.

        Returns:
            torch.Tensor : ``torch.tensor`` will be returned.
        """
        result_d = self.model_d(x)
        return result_d


@MODELS.register_module()
class DeblurGanV2Discriminator:
    """Defines the discriminator for DeblurGanv2 with the specified arguments..

    Args:
        model (Str): Type of the discriminator model
    """

    def __new__(cls, backbone, *args, **kwargs):
        if backbone == 'DoubleGan':
            return DoubleGan(*args, **kwargs)
        elif backbone == 'NoGan' or backbone == '':
            return super().__new__(cls)
        elif backbone == 'PatchGan':
            return PatchGan(*args, **kwargs)
        elif backbone == 'MultiScale':
            return MultiScale(*args, **kwargs)
        else:
            raise Exception('Discriminator model {} not found, '
                            'Please use the following models: '
                            '{}'.format(backbone, backbone_list))



# Copyright (c) OpenMMLab. All rights reserved.
from .singan import SinGAN
from .singan_discriminator import SinGANMultiScaleDiscriminator
from .singan_generator import SinGANMultiScaleGenerator

__all__ = [
    'SinGAN', 'SinGANMultiScaleDiscriminator', 'SinGANMultiScaleGenerator'
]



# Copyright (c) OpenMMLab. All rights reserved.
from .gca import GCA
from .gca_module import GCAModule
from .resgca_dec import ResGCADecoder, ResNetDec, ResShortcutDec
from .resgca_enc import ResGCAEncoder, ResNetEnc, ResShortcutEnc

__all__ = [
    'GCA', 'GCAModule', 'ResNetEnc', 'ResShortcutEnc', 'ResGCAEncoder',
    'ResNetDec', 'ResShortcutDec', 'ResGCADecoder'
]
